@@ -136,8 +136,32 @@ class _CompactContext(threading.local):
                     setattr(tree, field, subtree)
 
 
+@dataclass
+class _MutableContext(threading.local):
+    prev_mutable: tp.Optional[tp.Dict["Tree", bool]] = None
+
+    def __enter__(self):
+        global _MUTABLE_CONTEXT
+        self._old_context = _MUTABLE_CONTEXT
+        _MUTABLE_CONTEXT = self
+
+    def __exit__(self, *args):
+        global _MUTABLE_CONTEXT
+        _MUTABLE_CONTEXT = self._old_context
+
+    @contextmanager
+    def update(self, **kwargs):
+        fields = vars(self).copy()
+        fields.pop("_old_context", None)
+        fields.update(kwargs)
+
+        with _MutableContext(**fields):
+            yield
+
+
 _COMPACT_CONTEXT = _CompactContext()
 _CONTEXT = _Context()
+_MUTABLE_CONTEXT = _MutableContext()
 
 
 class FieldInfo:
@@ -158,8 +182,8 @@ class FieldInfo:
 
 
 class TreeMeta(ABCMeta):
-    def __call__(cls, *args, **kwargs) -> "Tree":
-        obj: Tree
+    def __call__(cls: tp.Type[T], *args, **kwargs) -> T:
+        obj: T
 
         if _COMPACT_CONTEXT.existing_subtrees is not None:
             if len(_COMPACT_CONTEXT.existing_subtrees) <= _COMPACT_CONTEXT.tree_idx:
@@ -173,35 +197,39 @@ class TreeMeta(ABCMeta):
             return obj
         else:
             obj = cls.__new__(cls)
-            obj = cls.construct(obj, *args, **kwargs)
+            with _make_mutable_single(obj):
+                obj = cls.construct(obj, *args, **kwargs)
 
         if _COMPACT_CONTEXT.new_subtrees is not None:
             _COMPACT_CONTEXT.new_subtrees.append(obj)
 
+        if _COMPACT_CONTEXT.in_compact and _COMPACT_CONTEXT.current_tree is not None:
+            _set_mutable(obj, _COMPACT_CONTEXT.current_tree._mutable)
+
+            if _MUTABLE_CONTEXT.prev_mutable is not None:
+                _MUTABLE_CONTEXT.prev_mutable[obj] = False
+
         return obj
 
     def construct(cls, obj: T, *args, **kwargs) -> T:
-        with _make_mutable_single(obj):
-            obj._field_metadata = obj._field_metadata.copy()
 
-            # set default fields
-            for field, default_factory in obj._factory_fields.items():
-                setattr(obj, field, default_factory())
+        obj._field_metadata = obj._field_metadata.copy()
 
-            for field, default_value in obj._default_field_values.items():
-                setattr(obj, field, default_value)
+        # set default fields
+        for field, default_factory in obj._factory_fields.items():
+            setattr(obj, field, default_factory())
 
-            # reset context before __init__ and add obj as current tree
-            with _CompactContext(current_tree=obj):
-                obj.__init__(*args, **kwargs)
+        for field, default_value in obj._default_field_values.items():
+            setattr(obj, field, default_value)
 
-            # auto-annotations
-            obj._update_local_metadata()
+        # reset context before __init__ and add obj as current tree
+        with _CompactContext(current_tree=obj):
+            obj.__init__(*args, **kwargs)
 
-            if _COMPACT_CONTEXT.current_tree is not None:
-                obj._mutable = _COMPACT_CONTEXT.current_tree._mutable
+        # auto-annotations
+        obj._update_local_metadata()
 
-            return obj
+        return obj
 
 
 class Tree(metaclass=TreeMeta):
@@ -466,22 +494,23 @@ def _make_mutable(obj: tp.Any):
     Context manager that makes the tree mutable.
     """
 
-    prev_mutable: tp.Dict[Tree, bool] = {}
-
     def _make_mutable_fn(a: Tree):
-        prev_mutable[a] = a._mutable
+        assert _MUTABLE_CONTEXT.prev_mutable is not None
+        _MUTABLE_CONTEXT.prev_mutable[a] = a._mutable
         # update __dict__ instead to avoid error during __setattr__
         a.__dict__["_mutable"] = True
 
     def _revert_mutable_fn(a: Tree):
+        assert _MUTABLE_CONTEXT.prev_mutable is not None
         # update __dict__ instead to avoid error during __setattr__
-        a.__dict__["_mutable"] = prev_mutable[a]
+        a.__dict__["_mutable"] = _MUTABLE_CONTEXT.prev_mutable[a]
 
-    try:
-        apply(_make_mutable_fn, obj, inplace=True)
-        yield
-    finally:
-        apply(_revert_mutable_fn, obj, inplace=True)
+    with _MUTABLE_CONTEXT.update(prev_mutable={}):
+        try:
+            apply(_make_mutable_fn, obj, inplace=True)
+            yield
+        finally:
+            apply(_revert_mutable_fn, obj, inplace=True)
 
 
 @contextmanager
@@ -502,3 +531,7 @@ def _make_mutable_single(*objs: Tree):
         # update __dict__ instead to avoid error during __setattr__
         for obj, _mutable in zip(objs, _mutables):
             obj.__dict__["_mutable"] = _mutable
+
+
+def _set_mutable(obj: Tree, value: bool):
+    obj.__dict__["_mutable"] = value
